"""
Utility to manage running commands in subprocesses.
Provides contextual information and higher-level monitoring
over the run commands.

@author: Baptiste Pestourie
@date: 20.11.2025
"""

from __future__ import annotations

import asyncio
import atexit
import io
import selectors
import signal
import subprocess
import time
import warnings
from asyncio import AbstractEventLoop
from dataclasses import dataclass, field
from functools import cache
from pathlib import Path
from subprocess import Popen
from threading import Lock, Thread
from typing import Callable, Self


@dataclass
class CommandContext:
    """
    Stores information about the execution of a command
    run in a different process.
    """

    args: list[str] = field(default_factory=list)
    cwd: Path = field(default_factory=Path.cwd)
    exit_code: int | None = None  # None means not completed
    stdout: list[str] = field(default_factory=list)
    stderr: list[str] = field(default_factory=list)
    start_time: float | None = None
    execution_time: float | None = None

    def __hash__(self) -> int:
        """
        unsafe hash, hashes by ID -
        thius object is mutable, and expected to be mutated after instanciation,
        so using fields for a hash is not a reliable option.
        """
        return id(self)

    @property
    def pending(self) -> bool:
        """
        Whether execution is still running.
        """
        return self.exit_code is None

    @property
    def success(self) -> bool:
        """
        Whether the execution succedded,
        i.e., exit code was 0.
        """
        return self.exit_code == 0

    @property
    def command(self) -> str:
        """
        The command that's getting run, in a single-line format.
        """
        return " ".join(self.args)

    def render(self) -> str:
        """
        Renders a log-friendly string version of this object.
        """
        lines: list[str] = []
        header_elements = [f"[{self.cwd}]", ">", self.command]
        match self.exit_code:
            case 0:
                status = "OK"
            case None:
                status = "PENDING"
            case _:
                status = "[FAILED]"

        header_elements.append(status)
        if self.execution_time:
            header_elements.append(f"[{self.execution_time:.2f}] s")
        lines.append(" ".join(header_elements))
        if self.stdout:
            lines.extend(["Stdout", "------", *self.stdout])
        if self.stderr:
            lines.extend(["Stderr", "------", *self.stderr])
        return "\n".join(lines)

    def __str__(self) -> str:
        """
        Shows the command and the folder from which it was called.
        """
        return f"[{self.cwd}] {self.command}"


class ProcessGarbageCollector(Thread):
    """
    A independant thread that takes care of cleaning up processes
    when timing out waiting for their completion.
    Sends first SIGINT, and falls back to SIGKILL if the process
    did not stop within `sigint_allowance_time`.

    Intended to be used mainly as a singleton-style, global shared instance,
    but safe to use as standalone, non-global object.
    Use `get_singleton` to get a reference to the singleton instance.
    """

    def __init__(self, sigint_allowance_time: float = 1.0) -> None:
        """
        Parameters
        ----------
        sigint_allowance_time
            How much time is given to the process to terminate after sending SIGINT.
            Once that time is reached, falls back to SIGKILL.
        """
        super().__init__(daemon=True)
        self._watched_processes: dict[CommandContext, Popen[str]] = {}
        self._loop: AbstractEventLoop | None = None
        self._lock = Lock()
        self._sigint_allowance_time = sigint_allowance_time

    @classmethod
    @cache
    def get_singleton(cls) -> Self:
        """
        Returns a single, shared global instance of that garbage collector.
        For most purposes this should be the favored way of using this object.
        Registers the cleanup function of that object on exit and starts
        the collection thread in the background.
        """
        global_instance = cls()
        global_instance.enable_cleanup_on_exit()
        global_instance.start()
        return global_instance

    def enable_cleanup_on_exit(self) -> None:
        """
        Registers the `shutdown_handler` of this object
        on interpreter termination
        """
        atexit.register(self.shutdown_handler)

    def interrupt(self, process: Popen[str], context: CommandContext) -> None:
        """
        Registers a process for graceful termination by sending SIGINT and monitoring it.
        """
        try:
            process.send_signal(signal.SIGINT)
        except ProcessLookupError:
            return  # Process already dead
        self._watched_processes[context] = process

        with self._lock:
            if (loop := self._loop) is not None:
                self._schedule_kill(loop, process, context)

    def run(self) -> None:
        """
        Main loop of the garbage collector to manage terminating processes.
        """
        loop = asyncio.new_event_loop()
        self._loop = loop
        asyncio.set_event_loop(loop)
        with self._lock:
            for context, proc in self._watched_processes.items():
                self._schedule_kill(loop, proc, context)
        loop.run_forever()

    def _schedule_kill(
        self, loop: AbstractEventLoop, process: Popen[str], context: CommandContext
    ) -> None:
        """
        Schedules process kill once the sigint allowance time is reached.
        """
        left_time = (
            self._sigint_allowance_time
            if context.start_time is None
            else context.start_time + self._sigint_allowance_time - time.time()
        )
        loop.call_soon_threadsafe(
            loop.call_later,
            left_time,
            self._kill_process,
            process,
            context,
        )

    def _update_context(self, context: CommandContext, exit_code: int) -> None:
        """
        Updates the context info with the execution time and exit code returned
        by the process.
        """
        if context.start_time is not None:
            context.execution_time = time.time() - context.start_time
        context.exit_code = exit_code

    def _kill_process(self, proc: Popen[str], context: CommandContext) -> None:
        """
        Calls SIKILL on the process `proc` if it's not dead yet.
        """
        self._refresh_state()
        if context not in self._watched_processes:
            return
        proc.kill()
        self._update_context(context, exit_code=0x80 - signal.SIGKILL)

    def _refresh_state(self) -> None:
        """
        Checks the status of all monitored processes and untrack the
        ones that have terminated already.
        """
        ctime = time.time()
        terminated_procs: list[CommandContext] = []
        for context, proc in self._watched_processes.items():
            if (exit_code := proc.poll()) is not None:
                terminated_procs.append(context)
                if exit_code != context.exit_code:
                    warnings.warn(
                        "[Process Garbage Collector]: we sent SIGINT, "
                        f"but process returned exit code {exit_code}"
                    )
                    context.exit_code = exit_code
            if context.start_time is not None:
                context.execution_time = ctime - context.start_time
        for ctx in terminated_procs:
            del self._watched_processes[ctx]

    def shutdown_handler(self) -> None:
        """
        Kills all monitored processes on interpreter shutdown.
        """
        with self._lock:
            for proc in self._watched_processes.values():
                if proc.poll() is None:
                    proc.kill()


def call_command(
    *args: str,
    timeout: float | None = None,
    printer: Callable[[str], None] | None = None,
    cwd: Path | str | None = None,
    process_gc: ProcessGarbageCollector | None = None,
    on_popen: Callable[[Popen[str]], None] | None = None,
) -> CommandContext:
    """
    Calls a command without blocking on IO, using selectors.
    It reads stdout and stderr as data comes in.
    It will leave early if the timeout is reached.
    The optional printer is called for each line received.

    Parameters
    ----------
    timeout
        If passed, wait for completion for a maximum of `timeout`
        and submit the process to the ProcessGarbageCollector when
        timing out.
        The process will be interrupted and later killed and it fails to
        terminate before the PGC allowance time.
    printer
        Optional, a function that receive each line from stderr/stdout
        from the proces.
    cwd
        Optional, the folder from which the process was called,
        assumes Path.cwd() by default.
    process_gc
        Optional, a custom ProcessGarbageCollector.
        Defaults to the global one.
    on_popen
        Optional, a callback to call after creating the Popen object.
        Should take the Popen object as unique argument.
    """
    start_time = time.time()
    current_cwd = Path(cwd) if cwd else Path.cwd()
    cmd = " ".join(args)
    proc = Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=current_cwd,
        text=True,  # Enable text mode
        bufsize=1,  # Request line buffering
        shell=True,
    )
    if on_popen is not None:
        on_popen(proc)
    should_interrupt = False
    stdout_lines: list[str] = []
    stderr_lines: list[str] = []

    sel = selectors.DefaultSelector()
    if proc.stdout:
        sel.register(proc.stdout, selectors.EVENT_READ, "stdout")
    if proc.stderr:
        sel.register(proc.stderr, selectors.EVENT_READ, "stderr")

    while sel.get_map():
        remaining_timeout = None
        if timeout is not None:
            elapsed_time = time.time() - start_time
            if elapsed_time >= timeout:
                break
            remaining_timeout = timeout - elapsed_time

        events = sel.select(timeout=remaining_timeout)

        if not events:
            # Timeout in select
            break

        for key, _ in events:
            # key.fileobj will be an io.TextIOWrapper because text=True
            assert isinstance(key.fileobj, io.TextIOWrapper)
            line = key.fileobj.readline()
            if not line:  # EOF for text streams is an empty string
                sel.unregister(key.fileobj)
                continue

            # readline() includes the newline, rstrip() to remove it before printing/storing
            processed_line = line.rstrip("\n")

            if printer:
                printer(processed_line)

            if key.data == "stdout":
                stdout_lines.append(processed_line)
            else:  # stderr
                stderr_lines.append(processed_line)

    final_exit_code = proc.poll()

    if final_exit_code is None:  # Process still running -> timeout occurred
        should_interrupt = True
        # Assume standard behavior for termination by SIGINT (Ctrl+C)
        final_exit_code = 0x80 + signal.SIGINT  # 130

    execution_time = time.time() - start_time

    context = CommandContext(
        args=list(args),
        cwd=current_cwd,
        exit_code=final_exit_code,
        stdout=stdout_lines,
        stderr=stderr_lines,
        start_time=start_time,
        execution_time=execution_time,
    )
    if should_interrupt:
        process_gc = process_gc or ProcessGarbageCollector.get_singleton()
        process_gc.interrupt(proc, context)
    return context



"""
Test suite for the C-extension compile tools.

@date: 27.05.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import importlib
import os
import platform
import shutil
import sysconfig
from contextlib import contextmanager
from pathlib import Path
from typing import Final, Generator, Literal, assert_never, cast, get_args

import pytest
from mypyc.build import mypycify
from setuptools import Extension

from smelt.compiler import (
    PYCONFIG_PATH,
    SupportedPlatforms,
    compile_extension,
    get_extension_suffix,
)

TEST_FOLDER: Final[Path] = Path(__file__).parent
EXTENSION_FOLDER = TEST_FOLDER / "extensions"
MODULE_FOLDER = TEST_FOLDER / "modules"


# All the extensions available in `extensions` folder
TestExtension = Literal["hello"]
TestModule = Literal["fib", "entrypoint"]

AVAILABLE_EXTENSIONS: list[str] = list(get_args(TestExtension))
AVAILABLE_MODULES: Final[list[str]] = list(get_args(TestModule))


@pytest.fixture(scope="session")
def local_platform_triple() -> str:
    """
    Returns
    -------
    str
        Platform triple of the device running this test suite
        Will be used to check some details of the output of native compilation
        tests, as their outputs are by definition platform dependant.
    """
    # harc-ocoding gnu
    system = platform.system().lower()
    match system:
        case "darwin":
            return f"{platform.machine()}-apple-darwin"

        case "windows":
            # TODO: one might actually use mingw32 on Windows
            # which would make this triple incorrect...
            # maybe use sysconfig.get_platform() ?
            return f"{platform.machine()}-windows-msvc"

        case "linux":
            return f"{platform.machine()}-linux-gnu"

        case _:
            raise RuntimeError(f"Running tests on unsupported OS: {system}")


def test_so_extension_suffix(local_platform_triple: str) -> None:
    """
    Checks that `get_extension_suffix` produces the same output
    as the sysconfig EXT_SUFFIX when used on the local platform
    """
    so_suffix = get_extension_suffix(local_platform_triple)
    assert so_suffix == sysconfig.get_config_var("EXT_SUFFIX")


@pytest.mark.parametrize(
    "python_version,platform_triple",
    [
        ["python3.12", "aarch64-linux-gnu"],
    ],
)
def test_pyconfig_headers_bundled(python_version: str, platform_triple: str) -> None:
    """
    Verifies that the pyconfig.h headers are properly bundled
    as package data
    """
    base_path = Path(PYCONFIG_PATH)
    pyconfig_path = base_path / platform_triple / python_version / "pyconfig.h"
    assert os.path.exists(pyconfig_path), f"Missing pyconfig.h @{pyconfig_path}"


@contextmanager
def build_temp_extension(
    ext_name: TestModule | TestExtension,
    crosscompile: SupportedPlatforms | None = None,
) -> Generator[str, None, None]:
    if ext_name in AVAILABLE_MODULES:
        cast(TestModule, ext_name)
        (extension,) = mypycify_module(ext_name)
        shared_lib_path = compile_extension(extension, crosscompile=crosscompile)
    else:
        shared_lib_path = compile_extension(
            get_extension_path(ext_name), crosscompile=crosscompile
        )
    try:
        yield shared_lib_path
    finally:
        if extensions_folder := os.environ.get("SAVE_EXTENSIONS", None):
            shutil.copy(shared_lib_path, extensions_folder)
        os.remove(shared_lib_path)


def mypycify_module(ext_name: TestModule) -> list[Extension]:
    """
    Builds a C extensioon out of a Python module using mypyc.
    """
    # Expecting to get one extension for mypyc runtime and one for the module
    return mypycify([str(get_module_path(ext_name))])


def get_extension_path(ext_name: TestExtension) -> Path:
    """
    Returns
    -------
    Path
        Full path to the requested extension
    """
    return EXTENSION_FOLDER / (ext_name + ".c")


def get_module_path(mod_name: TestModule) -> Path:
    """
    Returns
    -------
    Path
        Full path to the requested module
    """
    return MODULE_FOLDER / (mod_name + ".py")


@pytest.mark.parametrize("ext_name", AVAILABLE_EXTENSIONS)
def test_extensions_sanity(ext_name: TestExtension) -> None:
    """
    Checks if all available extensions have valid paths
    """
    assert os.path.exists(get_extension_path(ext_name))


@pytest.mark.parametrize("mod_name", AVAILABLE_MODULES)
def test_mypycify_sanity(mod_name: TestModule) -> None:
    """
    Checks if all available extensions have valid paths
    """
    assert len(mypycify_module(mod_name)) == 1


@pytest.mark.parametrize("ext_name", AVAILABLE_EXTENSIONS)
def test_compiler_builds_so(ext_name: TestExtension) -> None:
    """
    Verifies that the compiler is able to build a shared library
    """
    with build_temp_extension(ext_name) as shared_lib_path:
        assert os.path.exists(shared_lib_path)
        assert shared_lib_path.endswith(".so")
    assert not os.path.exists(shared_lib_path), (
        "`build_temp_extension` fixture did not clean-up properly"
    )


@pytest.mark.parametrize("ext_name", AVAILABLE_EXTENSIONS)
@pytest.mark.parametrize("platform", [SupportedPlatforms.AARCH64_LINUX])
def test_compiler_crosscompiled_so(
    ext_name: TestExtension, platform: SupportedPlatforms
) -> None:
    """
    Verifies that the compiler is able to build a shared library for a foreign platform.
    Does not verify that the shared library works - this requires tooling beyond
    what a isolated Python test session can do.
    """
    with build_temp_extension(ext_name, crosscompile=platform) as shared_lib_path:
        assert os.path.exists(shared_lib_path)
        assert shared_lib_path.endswith(".so")
    assert not os.path.exists(shared_lib_path), (
        "`build_temp_extension` fixture did not clean-up properly"
    )


@pytest.mark.parametrize("ext_name", AVAILABLE_EXTENSIONS)
def test_built_so(ext_name: TestExtension) -> None:
    """
    Verifies that the built shared library works
    """
    with build_temp_extension(ext_name):
        ext_mod = importlib.import_module(ext_name)

    match ext_name:
        case "hello":
            hello_func = ext_mod.hello
            assert hello_func() == "Hello World!"

        case _ as unreachable:
            assert_never(unreachable)


@pytest.mark.parametrize("mod_name", AVAILABLE_MODULES)
def test_compiler_compliant_with_mypyc(mod_name: TestModule) -> None:
    with build_temp_extension(mod_name) as shared_lib_path:
        assert os.path.exists(shared_lib_path)
        assert shared_lib_path.endswith(".so")
    assert not os.path.exists(shared_lib_path)


@pytest.mark.parametrize("mod_name", AVAILABLE_MODULES)
def test_compiler_built_mypyc(mod_name: TestModule) -> None:
    with build_temp_extension(mod_name) as shared_lib_path:
        assert os.path.exists(shared_lib_path)
        assert shared_lib_path.endswith(".so")
        ext_mod = importlib.import_module(mod_name)

        match mod_name:
            case "fib":
                fib = ext_mod.fib
                assert fib(10) == 55

            case "entrypoint":
                import subprocess
                import sys

                p = subprocess.run(
                    [sys.executable, "-c", "import " + mod_name, "10"],
                    check=True,
                )
                # .so modules cannot be run as main,
                # so just checking here that it imports without errors
                assert p.returncode == 0, (
                    "Entrypoint module did not run successfully.: " + shared_lib_path
                )

            case _ as unreachable:
                assert_never(unreachable)



"""
Verifies some properties from auto-generated packages using Smelt
build hook.
@date: 11.12.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations
import sys
import pytest
from types import ModuleType
from pytest import FixtureRequest
import importlib

from smelt.process import call_command

USAGE_HINT = """\
These tests are primarily meant for nox and should be run by nox directly.
You can still use this utility to test example packages manually by passing
the package name explicitly with --package {package_name} when calling pytest.
"""

PKG_NOT_FOUND_HINT = """\
Failed to import {package_name}.
If running without nox and using --package,
make sure that the target package `{package_name}` is installed in your venv.
If using nox, check the nox session parameters. 
"""


@pytest.fixture
def package_name(request: FixtureRequest) -> str | None:
    return request.config.getoption("--package")


def try_import_cli_module(package_name: str | None) -> ModuleType:
    if package_name is None:
        assert False, USAGE_HINT
    try:
        return importlib.import_module(f"{package_name}.cli")
    except ImportError:
        assert False, PKG_NOT_FOUND_HINT.format(package_name=package_name)


def test_package_import_cli(package_name: str | None) -> None:
    cli_mod = try_import_cli_module(package_name)
    # verifying that the entrypoint function `cli` is in there
    assert hasattr(cli_mod, "compute_fib")


def test_cli_mod_returns_correct_value(package_name: str | None) -> None:
    cli_mod = try_import_cli_module(package_name)
    # verifying that the entrypoint function `cli` is in there
    ctx = call_command("compute-fib", "10", timeout=1.0, printer=print)
    assert ctx.exit_code == 0
    assert len(ctx.stdout) == 1
    assert ctx.stdout[0].strip() == "55"



"""
Defines a distutils compiler based on Zig.

@date: 27.05.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import contextlib
import logging
import os
import shutil
import subprocess
import sys
import sysconfig
import tempfile
import warnings
from enum import StrEnum
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Final

from distutils.compilers.C.unix import Compiler
from setuptools import Extension

from smelt.utils import (
    ImportPath,
    PathSolver,
    SmeltError,
    assert_path_exists,
    get_extension_suffix,
    PathExists,
    path_exists,
)
from smelt.process import call_command

if TYPE_CHECKING:
    from os import PathLike

_SMELT_ROOT: Final[str] = os.path.dirname(__file__)
PYCONFIG_PATH: Final[str] = os.path.join(_SMELT_ROOT, "pyconfig")

_logger = logging.getLogger(__name__)


class SupportedPlatforms(StrEnum):
    """
    All the target platforms supported from cross-compilation.
    Value of the enum corresponds the platform name as expected by Zig compiler.
    """

    # TODO: parametrize OS
    AARCH64_LINUX = "aarch64-linux"
    ARMV7L_LINUX = "arm-linux-gnueabihf"
    X86_64_LINUX = "x86_64-linux"
    # TODO: add more

    def get_triple_name(self) -> str:
        """
        Returns the "triple" platform name <arch>-<os>-<libc>
        as used by Python for this target.
        Note: automatically assumed libc here, as there's no support currently for other options.
        """
        # Note: Python can be built for multiple LibCs:
        # gnu, musl, android...
        # Currently hard-coding LibC, which would be the choices for 95%+ projects
        # out there. Other libC might be considered later
        if self == SupportedPlatforms.ARMV7L_LINUX:
            return self.value
        return self.value + "-gnu"


class ZigCompiler(Compiler):
    """
    Base class for zig compiler.
    Sets the C/C++ compiler exe to zig cc and zig c++ and
    add .zig files to the list of accepted extensions.
    """

    zig_base_path: ClassVar[list[str]] = [sys.executable, "-m", "ziglang"]
    # Expanding to add .zig files
    src_extensions: ClassVar[list[str]] = Compiler.src_extensions + [".zig"]

    executables = {
        "preprocessor": None,
        "compiler": [*zig_base_path, "cc"],
        "compiler_so": [*zig_base_path, "cc"],
        "compiler_cxx": [*zig_base_path, "c++"],
        "compiler_so_cxx": [*zig_base_path, "c++"],
        "linker_so": [*zig_base_path, "cc", "-shared"],
        "linker_so_cxx": [*zig_base_path, "c++", "-shared"],
        "linker_exe": [*zig_base_path, "cc"],
        "linker_exe_cxx": [*zig_base_path, "c++", "-shared"],
        "archiver": ["ar", "-cr"],
        "ranlib": None,
    }


def zig_build_lib(
    name: str,
    object_files: list[str],
    crosscompile: SupportedPlatforms | None = None,
) -> str:
    """
    Builds a shared library using Zig's native interface (build-lib)
    instead of zig cc.

    Parameters
    ----------
    name: str
        The name of the library to build, without extension.

    object_files: list[str]
        List of object files to include in the shared library.

    crosscompile: SupportedPlatforms | None
        The target platform for cross-compilation.
        If None, builds for the current platform.

    Returns
    -------
    str
        The path to the built shared library.
    """
    cmd = ["python", "-m", "ziglang", "build-lib"]
    if crosscompile is not None:
        cmd.extend(["-target", crosscompile.value])
    # we need position independant code
    cmd.append("-dynamic")
    # as we'll wrapp this into a shared library later
    cmd.append("-fPIC")
    # Python itself is only linked at runtime,
    # so we need to allow undefined symbols
    cmd.append("-fallow-shlib-undefined")
    cmd.extend(["--name", name])
    cmd.extend(object_files)

    _logger.info("Running zig build-lib: \n%s", " ".join(cmd))
    subprocess.run(cmd)
    # Note: zig build-lib will produce a file named lib{name}.so

    if crosscompile is not None:
        suffix = get_extension_suffix(crosscompile.get_triple_name())
    else:
        suffix = sysconfig.get_config_var("EXT_SUFFIX")
    # copying
    dest_name = f"{name}{suffix}"
    # copying the shared library to the expected name
    shutil.copy(f"lib{name}.so", dest_name)
    return dest_name


def zig_build_exe(
    name: str,
    object_files: list[str],
    crosscompile: SupportedPlatforms | None = None,
) -> str:
    """
    Builds a shared library using Zig's native interface (build-lib)
    instead of zig cc.

    Parameters
    ----------
    name: str
        The name of the library to build, without extension.

    object_files: list[str]
        List of object files to include in the shared library.

    crosscompile: SupportedPlatforms | None
        The target platform for cross-compilation.
        If None, builds for the current platform.

    Returns
    -------
    str
        The path to the built shared library.
    """
    cmd = ["python", "-m", "ziglang", "build-exe"]
    if crosscompile is not None:
        cmd.extend(["-target", crosscompile.value])
    # we need position independant code
    # Python itself is only linked at runtime,
    # so we need to allow undefined symbols
    cmd.extend(["--name", name])
    cmd.extend(object_files)

    _logger.info("Running zig build-lib: \n%s", " ".join(cmd))
    subprocess.run(cmd)
    # Note: zig build-lib will produce a file named lib{name}.so

    if crosscompile is not None:
        suffix = get_extension_suffix(crosscompile.get_triple_name())
    else:
        suffix = sysconfig.get_config_var("EXT_SUFFIX")
    # copying
    dest_name = f"{name}{suffix}"
    # copying the shared library to the expected name
    # shutil.copy(f"lib{name}.so", dest_name)
    return dest_name


def compile_zig_module(
    name: str,
    folder: PathExists,
    import_path: ImportPath,
    path_solver: PathSolver | None = None,
    flags: list[str] | None = None,
) -> PathExists:
    path_solver = path_solver or PathSolver()
    flags = flags or []
    with contextlib.chdir(folder):
        call_command("zig", "build", *flags)
        # TODO windows
        lib_path = Path.cwd() / "zig-out" / "lib" / (name + ".so")
    if not path_exists(lib_path):
        raise SmeltError(
            f"Ran `zig build` successfully, but no library `{lib_path}` was found afterwards"
            "Check that the name of the project is properly configured, as well as your build.zig"
        )
    # TODO crosscompile
    suffix = sysconfig.get_config_var("EXT_SUFFIX")
    target_path = path_solver.resolve_import_path(
        import_path, file_extension=suffix, should_exist=False
    )
    shutil.move(
        lib_path,
        target_path,
    )
    assert path_exists(target_path)
    return target_path


def compile_extension(
    extension: Path | str | Extension,
    compiler: Compiler | None = None,
    dest_folder: PathLike[str] | None = None,
    crosscompile: SupportedPlatforms | None = None,
    use_zig_native_interface: bool = False,
) -> PathExists:
    """
    Standalone function compiling a low-level extension (C, C++ or Zig)
    into a shared library.

    Parameters
    ----------
    extension_path: Path | str | Extension
        Path to the source file to compile,
        or a pre-built Extension object

    compiler: Compiler | None
        The compiler to use,
        spawns a ZigCompiler if omitted

    dest_folder: PathLike[str]
        The folder in which to place the built shared library.
        Defaults to cwd.
    """
    compiler = compiler or ZigCompiler()
    include_dirs = [sysconfig.get_path("include"), sysconfig.get_path("platinclude")]
    library_dirs = [sysconfig.get_config_var("LIBDIR")]

    if isinstance(extension, (str, Path)):
        if not os.path.exists(extension):
            raise FileNotFoundError(f"Extension does not exists: {extension}")

        # building an extension object for a single source file
        extension = Path(extension)
        ext_name = extension.name

        if extension.suffix not in compiler.src_extensions:
            raise ValueError(
                f"Unsupported extension: {extension.suffix} "
                f"Supported values: {','.join(compiler.src_extensions)}"
            )
        extension_obj = Extension(
            name=extension.name.replace(extension.suffix, ""),
            sources=[
                str(extension),
            ],
        )
    else:
        extension_obj = extension
        ext_name = extension.name

    # Compile the C file
    extra_preargs: list[str] = []
    if crosscompile is not None:
        # TODO: generate/obtain pyconfig.h for the target platform
        warnings.warn(
            "Support for cross-compiling is experimental.\n"
            "Do not assume stability from the built artifacts"
        )
        extra_preargs.append(f"--target={crosscompile.value}")
        # adding pyconfig
        include_dirs.append(PYCONFIG_PATH)
        so_suffix = get_extension_suffix(crosscompile.get_triple_name())
    else:
        so_suffix = sysconfig.get_config_var("EXT_SUFFIX")

    with tempfile.TemporaryDirectory() as build_folder:
        # TODO: investigate the pure setuptools alternative
        # as the distutils compiler is deprecated
        objects = compiler.compile(
            sources=extension_obj.sources,
            output_dir=build_folder,
            include_dirs=include_dirs + extension_obj.include_dirs,
            extra_preargs=extra_preargs,
            extra_postargs=extension_obj.extra_compile_args or [],
            macros=extension_obj.define_macros,
        )

        # Link it into a shared object
        ext_name = extension_obj.name + so_suffix

        output_dir = dest_folder or "."

        if not use_zig_native_interface:
            compiler.link_shared_object(
                objects,
                ext_name,
                output_dir=str(output_dir),
                libraries=extension_obj.libraries,
                library_dirs=extension_obj.library_dirs + library_dirs,
                runtime_library_dirs=extension_obj.runtime_library_dirs,
                extra_preargs=extra_preargs,
            )
        else:
            zig_build_lib(extension.name, objects, crosscompile=crosscompile)
    so_path = os.path.join(output_dir, ext_name)
    return assert_path_exists(so_path)



import time

import click
from minimal.fib import fib as fib_mypyc
from minimal.fib_cython import fibx as fib_cython


def fib_pure_python(n: int) -> int:
    if n <= 1:
        return n
    else:
        return fib_pure_python(n - 2) + fib_pure_python(n - 1)


@click.group()
def minimal(): ...


@minimal.command()
def say_hello():
    from minimal.hello import hello

    hello()


@minimal.command()
@click.argument("n", type=int)
def compute_fib(n: int):
    for func in [fib_pure_python, fib_mypyc, fib_cython]:
        start_time = time.time()
        click.echo(f"Computing fibonacci of {n} with {func.__name__}")
        click.echo(func(n))
        click.echo(f"Computation took {time.time() - start_time}")


if __name__ == "__main__":
    minimal()



import time

import click

from hatchdemo.fib import fib as fib_mypyc
from hatchdemo.fib_purepython import fib as fib_purepython
from hatchdemo.fib_cython import fib as fib_cython


@click.group()
def hatchdemo(): ...


@hatchdemo.command()
def say_hello():
    from hatchdemo.hello import hello

    print(hello())


@hatchdemo.command()
@click.argument("n", type=int)
def compute_fib(n: int):
    start_time = time.time()
    click.echo(f"Computing fibonacci of {n} [pure python]")
    purepy_result = fib_purepython(n)
    click.echo(f"Computation took {time.time() - start_time} s")

    start_time = time.time()
    click.echo(f"Computing fibonacci of {n} [mypyc]")
    mypyc_result = fib_mypyc(n)
    click.echo(f"Computation took {time.time() - start_time} s")

    start_time = time.time()
    click.echo(f"Computing fibonacci of {n} [cython]")
    cython_result = fib_cython(n)
    click.echo(f"Computation took {time.time() - start_time} s")

    assert purepy_result == mypyc_result
    assert purepy_result == cython_result
    click.echo(f"Result is {purepy_result}, by the way")


if __name__ == "__main__":
    hatchdemo()



import signal
import sys
import time
from pathlib import Path

import pytest

from smelt.process import ProcessGarbageCollector, call_command

# The command echoes 3 numbers with a 100ms delay between each.
# Total execution time is ~300ms.
CMD = "sleep .1; echo 1; sleep .1; echo 2; sleep .1; echo 3"

UNSTOPPABLE_SCRIPT = """
import signal
import time
import sys

def handle_sigint(sig, frame):
    print("SIGINT ignored", flush=True)

signal.signal(signal.SIGINT, handle_sigint)

print("started", flush=True)
while True:
    time.sleep(1)
"""


def test_call_command_success() -> None:
    """
    Tests the successful execution of a command, checking exit code and stdout.
    The command should run to completion.
    """
    result = call_command(CMD)

    assert result.success is True
    assert result.exit_code == 0
    assert result.stdout == ["1", "2", "3"]
    assert result.stderr == []
    assert result.execution_time is not None
    # Check that it took at least 300ms.
    assert result.execution_time >= 0.3


def test_call_command_printer_timing() -> None:
    """
    Tests that the printer function is called progressively as output is generated,
    not all at once at the end.
    """
    received_lines_with_ts = []

    def printer_with_timestamp(line: str) -> None:
        received_lines_with_ts.append((line, time.time()))

    result = call_command(CMD, printer=printer_with_timestamp)

    assert result.exit_code == 0

    received_lines = [line for line, ts in received_lines_with_ts]
    assert received_lines == ["1", "2", "3"]

    # Verify that the prints happened at roughly 100ms intervals.
    timestamps = [ts for line, ts in received_lines_with_ts]

    # Check deltas between prints. Allow a generous range to avoid flaky tests.
    delta1 = timestamps[1] - timestamps[0]
    delta2 = timestamps[2] - timestamps[1]

    # These should be around 0.1 seconds. We check for a wide range [0.08, 0.25]
    # which is equivalent to 0.165 +/- 0.085
    assert delta1 == pytest.approx(0.165, abs=0.085)
    assert delta2 == pytest.approx(0.165, abs=0.085)


def test_call_command_timeout() -> None:
    """
    Tests that a timeout interrupts the command, resulting in a non-zero exit code
    and partial output.
    """
    # Total command time is > 0.3s. A timeout of 0.15s should cut it off after the first echo.
    timeout = 0.15

    result = call_command(CMD, timeout=timeout)
    assert result.success is False
    # On Unix, a process terminated by SIGINT (like Ctrl+C) often returns 130.
    assert result.exit_code == 130

    # Only the first number should be in stdout.
    assert result.stdout == ["1"]
    assert result.stderr == []

    assert result.execution_time is not None
    # Check that the function returned shortly after the timeout was hit.
    # We expect the execution time to be in the range [timeout, timeout + 0.1s].
    # This is equivalent to timeout + 0.05s +/- 0.05s
    assert result.execution_time == pytest.approx(timeout + 0.05, abs=0.05)


def test_gc_sigkill_fallback(tmp_path: Path) -> None:
    """
    Tests that the ProcessGarbageCollector falls back to SIGKILL if a process
    ignores SIGINT.
    """

    script_path = tmp_path / "ignore_sigint.py"
    script_path.write_text(UNSTOPPABLE_SCRIPT)

    procs = []

    timeout = sigint_allowance = 0.2
    # 2. Create a custom GC with a short allowance time for a fast test
    custom_gc = ProcessGarbageCollector(sigint_allowance_time=sigint_allowance)
    custom_gc.start()

    # 3. Run the command that will time out
    command_to_run = f"{sys.executable} {script_path}"
    result = call_command(
        command_to_run,
        timeout=timeout,
        process_gc=custom_gc,
        on_popen=procs.append,
    )

    # 4. Verify the immediate result from call_command
    assert result.success is False
    assert result.exit_code == 130  # Assumed SIGINT exit code
    assert result.stdout == ["started"]

    # 5. Wait for the GC to do its work (SIGINT allowance + SIGKILL)
    time.sleep(sigint_allowance + 0.1)

    # 6. Verify that the process was eventually SIGKILLed
    assert len(procs) == 1
    process = procs[0]

    # After being killed, poll() will return the exit code.
    # SIGKILL is signal 9, so the return code should be -9.
    assert process.poll() == -signal.SIGKILL


def test_gc_atexit_handler(tmp_path: Path) -> None:
    """
    Checks that the process garbage coillector kills everything immediately
    when the interpreter shutdown handler is called.
    In real conditions, this would be injected by atexit.
    """

    script_path = tmp_path / "ignore_sigint.py"
    script_path.write_text(UNSTOPPABLE_SCRIPT)

    procs = []

    # 2. Create a custom GC with a short allowance time for a fast test
    custom_gc = ProcessGarbageCollector()
    custom_gc.start()

    # 3. Run the command that will time out
    command_to_run = f"{sys.executable} {script_path}"
    result = call_command(
        command_to_run,
        timeout=0.1,
        process_gc=custom_gc,
        on_popen=procs.append,
    )

    # 4. Verify the immediate result from call_command
    assert result.success is False
    assert result.exit_code == 130  # Assumed SIGINT exit code
    assert result.stdout == ["started"]

    # 6. Verify that the process was eventually SIGKILLed
    assert len(procs) == 1
    process = procs[0]

    assert process.poll() == None
    # Simulating interpreter exit
    custom_gc.shutdown_handler()
    # process should stop now
    assert process.wait(0.1)
    assert process.poll() == -signal.SIGKILL



from __future__ import annotations
from typing import TYPE_CHECKING


from smelt.config import (
    NuitkaModule,
    MypycModule,
    CythonExtension,
    SmeltConfig,
    TomlData,
    build_datacls_from_toml,
)


import pytest

from smelt.utils import SmeltConfigError, TomlData

if TYPE_CHECKING:
    from _typeshed import DataclassInstance


@pytest.mark.parametrize("ConfigCls", [NuitkaModule, MypycModule])
@pytest.mark.parametrize(
    ["toml_data", "valid"],
    [
        ({"import_path": "a.b.c"}, True),
        ({"import_path": "a.$.c"}, False),
        ({"import_path": "a.$./"}, False),
        ({"import_path": ""}, False),
    ],
)
def test_config_conversion_import_path(
    ConfigCls: type[NuitkaModule | MypycModule],
    toml_data: TomlData,
    valid: bool,
) -> None:
    if valid:
        datacls = build_datacls_from_toml(ConfigCls, toml_data=toml_data)
        assert datacls.import_path == toml_data["import_path"]
    else:
        with pytest.raises(SmeltConfigError):
            build_datacls_from_toml(ConfigCls, toml_data=toml_data)



"""
A utility to provide convenience functions around managing
global/shared contexts within the scope of an application.
This implementation is designed to allow contexts to be optional
by design - i.e., they would only be used if the entrypoint application
has allowed them in the first place.
That allows to make the mutable shared state an opt-in thing,
and for library-style usage, these contexts should remain transparent
as long as they are not enabled.

Enable contexts with `enable_global_context()`.
use `get_context(<name>)` to get a global context instance
registered at the given `name`.
Create one under that name with `create_context(name)`.

Contexts objects themselves are defined by the `Context` protocol.
They should provid a `render` method that allows exporting them to a report.

@date: 27.11.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Protocol, overload


class Context(Protocol):
    """
    Base protocol for context objects.
    """

    def render(self) -> str: ...


@dataclass
class GlobalContext:
    """
    The main object manging all other contexts.
    Manages mainly two types of context objects:
    * Persistent contexts, that are stored under a given name
    * Traces, that contain information about a given operation in history,
    and are stored in an ordered way.
    """

    persistent_contexts: dict[str, Context] = field(default_factory=dict)
    traces: list[Context] = field(default_factory=list)

    def add_trace(self, trace: Context) -> None:
        """
        Stores a new trace in the global context.
        Added traces will be rendered in order of addition.
        """
        self.traces.append(trace)

    def clear_traces(self) -> None:
        """
        Clears all stored traces.
        """
        self.traces.clear()

    def create_context(self, name: str, context: Context) -> None:
        """
        Creates persistent context under the given name.
        """
        self.persistent_contexts[name] = context

    def get_context(self, name: str) -> Context | None:
        """
        Retrieves the persistent context stored under the given name.
        Returns None if no such context exists.
        """
        return self.persistent_contexts[name]

    def render(self) -> str:
        """
        Renders all persistent contexts and traces into a report.
        """
        lines: list[str] = []
        if self.persistent_contexts:
            lines.append("General context\n------------")
        for context_name, context in self.persistent_contexts.items():
            lines.append(context_name)
            lines.append("-" * len(context_name))
            lines.append(context.render())
        if self.traces:
            lines.append("Trace\n-------------")
        for trace in self.traces:
            lines.append("> " + trace.render())
        return "\n".join(lines)


@dataclass
class PathTrace:
    """
    Path from which we are operating.
    """

    cwd: Path = field(default_factory=Path.cwd)


@dataclass
class TaskTrace:
    """
    A generic context that takes abitrary string elements.
    Can be used to explain what part of the business logic is being
    executed at a given time in a given sequence, without any specific
    expectation on the format.
    """

    elements: list[str | Context] = field(default_factory=list)

    def comment(self, text: str) -> None:
        self.elements.append(text)

    def add_sub_context(self, context: Context) -> None:
        self.elements.append(context)

    def render(self) -> str:
        rendered_elements = [
            element if isinstance(element, str) else element.render()
            for element in self.elements
        ]
        return "\n".join(rendered_elements)


_GLOBAL_CONTEXT: None | GlobalContext = None


@overload
def get_context(name: None = None) -> GlobalContext | None: ...


@overload
def get_context(name: str) -> Context | None: ...


def get_context(name: str | None = None) -> Context | None:
    if name is None:
        return _GLOBAL_CONTEXT
    global_context = get_context()
    if global_context is None:
        return None
    return global_context.get_context(name)


def is_global_context_enabled() -> bool:
    """
    Whether contexts are enabled globally.
    """
    return _GLOBAL_CONTEXT is not None


def enable_global_context() -> None:
    """
    Creates the gloabl context and enables context operations
    within this interpreter scope.
    """
    global _GLOBAL_CONTEXT
    if _GLOBAL_CONTEXT is None:
        _GLOBAL_CONTEXT = GlobalContext()


def reset_contexts() -> None:
    """
    Resets all context.
    The global context is re-created,
    meaning all stored traces and persistent context will be lost.
    """
    global _GLOBAL_CONTEXT
    _GLOBAL_CONTEXT = None
    _GLOBAL_CONTEXT = GlobalContext()


def clear_contexts() -> None:
    """
    Clears all contexts, and disables context operations.
    """
    global _GLOBAL_CONTEXT
    _GLOBAL_CONTEXT = None


def create_context_if_enabled[Ctx: Context](
    name: str, context: type[Ctx] | Ctx
) -> Ctx | None:
    global_ctx = get_context()
    if global_ctx is None:
        return None
    if isinstance(context, type):
        context = context()

    global_ctx.create_context(name, context)
    return context



"""
Common utilities for this package.

@date: 30.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import importlib.util
import importlib
import keyword
import logging
import os
import shutil
import sys
import sysconfig
import tempfile
import warnings
from contextlib import contextmanager
from dataclasses import dataclass, field, fields
from enum import Enum, auto
from pathlib import Path
from types import ModuleType
from typing import (
    Generator,
    Literal,
    TypeGuard,
    NamedTuple,
    NewType,
    Self,
    assert_never,
    cast,
    overload,
)

from setuptools import Extension

from smelt.context import get_context

_logger = logging.getLogger(__name__)

PathExists = NewType("PathExists", Path)
ModuleName = NewType("ModuleName", str)
type _TomlData = str | list[_TomlData] | dict[str, _TomlData]
type TomlData = dict[str, _TomlData]


# --- Errors ---
class SmeltError(Exception):
    """
    Base exception class for Smelt errors
    """


class SmeltConfigError(SmeltError):
    """
    Raised when running conversions for config loading.
    """


class SmeltFileNotFoundError(SmeltError):
    """
    Raised when a file referenced by config is not found.
    """


class SmeltAlreadyNativeModule(SmeltError):
    """
    Raised when trying to convert to native an already native module.
    """


def path_exists(path: Path) -> TypeGuard[PathExists]:
    """
    Type magic.
    Allows representing existence of a given path in the type system.
    """
    return path.exists()


def assert_path_exists(path: str | Path) -> PathExists:
    path_obj = Path(path)
    if not path_exists(path_obj):
        raise SmeltConfigError(f"Following file/folder does not exist: {path}")
    return path_obj


def is_valid_import_path(import_path: str) -> TypeGuard[ImportPath]:
    """
    Checks if `import_path` is a valid import path in Python.
    """
    if not isinstance(import_path, str) or not import_path:
        return False

    parts = import_path.split(".")

    for part in parts:
        if not part.isidentifier():
            return False
        if keyword.iskeyword(part):
            return False
    return True


def assert_is_valid_import_path(import_path: str) -> ImportPath:
    if not is_valid_import_path(import_path):
        raise SmeltConfigError(f"Invalid python import path: {import_path}")
    return import_path


def is_valid_module_name(name: str) -> TypeGuard[ModuleName]:
    """
    Checks that `name` is a valid module name in Python.
    """
    if not is_valid_import_path(name):
        return False
    return "." not in name


def get_module_name(import_path: ImportPath) -> ModuleName:
    """
    Extracts the module name at the stem from `import_path`.
    """
    modname = import_path.split(".")[-1]
    assert is_valid_module_name(modname)
    return modname


class SmeltMissingModule(SmeltError): ...


def get_extension_suffix(target_triple: str) -> str:
    """
    Generate the C extension module filename.

    Parameters
    ----------
    target_triple: str
        The target triple, e.g., 'aarch64-linux-gnu'.

    Returns
    -------
    str
        The extension filename, e.g., '.cpython-312-aarch64-linux-gnu.so'
    """
    major = sys.version_info.major
    minor = sys.version_info.minor
    return f".cpython-{major}{minor}-{target_triple}.so"


def locate_module_by_import_path(import_path: ImportPath) -> PathExists:
    try:
        spec = importlib.util.find_spec(import_path)
    except ModuleNotFoundError as exc:
        # package not found
        raise SmeltMissingModule(
            f"{import_path} refers to some unknown packages in your environment. Is it installed ?"
        ) from exc

    if spec is None:  # module not found
        raise SmeltMissingModule(f"Module {import_path} not found. Is it installed ?")

    origin = spec.origin
    if origin is None:
        # PEP 420: path can be valid import without having an origin
        # if it refers to a namespace package,
        # as it might be found in multiple places.
        raise SmeltConfigError(
            f"Could not locate {import_path}. "
            "You might be referring to a namespace package, which is not a valid module to import"
        )
    module_path = Path(origin)
    assert path_exists(module_path), (
        "importlib.util.find_spec returned an invalid system path"
    )
    return module_path


@contextmanager
def import_shadowed_module(path: str) -> Generator[ModuleType, None, None]:
    """
    Allows importting a .py module shadowed by its .so counterpart
    (typically generated by mypy).
    .so python modules (generally, compiled C extensions) have higher priority on import
    over .py modules. This is is done so that module can provide optional C extensions
    with a pure Python implementation as fallback (e.g., if the interpreter is not CPython,
    or if the C extension cannot be compiled for the target platform)

    This function temporarily REMOVES the .so and proceeds to import the shadowed.
    It is a context manager and will restore the removed .so on exit,
    unless a new version has been built in the meantime (e.g., if the import
    was required as part of a recompile process)
    """
    with tempfile.TemporaryDirectory() as tmp:
        try:
            mod = importlib.import_module(path)
        except ImportError as exc:
            msg = f"Failed to import {path} while trying to mypycify"
            raise SmeltMissingModule(msg) from exc
        mod_path = mod.__file__
        assert mod_path is not None, f"Cannot module to mypycify: {path}"
        if mod_path.endswith((".py", ".pyc")):
            yield mod
        else:
            backup_path = os.path.join(tmp, os.path.basename(mod_path))
            shutil.copy(mod_path, backup_path)
            _logger.debug("Removing %s...", mod_path)
            os.remove(mod_path)
            os.sync()

            try:
                # using reload as otherwise importlib will use the cache
                # and re-import the now deleted .so
                mod = importlib.reload(mod)
                assert mod.__file__ is not None and not (mod.__file__.endswith(".so"))
                yield mod
            except ImportError as exc:
                msg = f"Failed to import {path} while trying to mypycify"
                raise SmeltMissingModule(msg) from exc


ImportPath = NewType("ImportPath", str)
FsPath = NewType("FsPath", str)


class ModpathType(Enum):
    """
    How locations of modules are passed:

    `IMPORT` means the module path is given as it would be used
    in a import statement.
    `FS` means the module is passed as a relative path.
    """

    IMPORT = auto()
    FS = auto()


@dataclass
class PathResolutionTrace:
    """
    Stores information about converting an import path to a filesystem path
    or vice-versa.
    """

    module_path: str
    import_path: str
    resolution_type: ModpathType

    def render(self) -> str:
        match self.resolution_type:
            case ModpathType.FS:
                return (
                    f"Resolving filesystem location of `{self.module_path}`:"
                    f"found `{self.import_path}`"
                )
            case ModpathType.IMPORT:
                return (
                    f"Resolving import path of file `{self.import_path}`:"
                    f"found `{self.module_path}`"
                )


def get_modpath_type(path: str) -> ModpathType:
    """
    Detects whether the passed type of the given `path`.
    """
    # TODO: stricter checks
    if path.endswith(".py") or path.endswith(".pyx"):
        return ModpathType.FS
    return ModpathType.IMPORT


class PackageLayout(Enum):
    """
    Which standard folder layout a given package is using.
    """

    SRC = auto()
    EPONYM = auto()


def detect_package_layout(package_name: str, package_root: str = ".") -> PackageLayout:
    """
    Tries detecting the package layout based on the file tree.
    """
    root = Path(package_root)
    has_src = (root / "src").exists()
    has_eponym_folder = (root / package_name).exists()

    if not has_src and not has_eponym_folder:
        raise RuntimeError(
            f"No `src` or `{package_name}` found cannot detect package layout"
        )
    if has_eponym_folder:
        if has_src:
            warnings.warn(
                "Package has both a `src` and a {package_name} folder at the root. "
                f"Assuming {package_name} contains source code, but that migth be wrong"
            )
            return PackageLayout.EPONYM

    if not (root / "src" / package_name).exists():
        raise RuntimeError(
            f"No `{package_name}` found inside `src` folder. "
            "If `src` is used as a substitute for your package name",
            "you need to configure that explicitly.",
        )
    return PackageLayout.SRC


def convert_to_path(import_path: ImportPath, file_extension: str = ".py") -> Path:
    *packages, module = import_path.split(".")
    module += file_extension
    return Path(*packages, module)


@overload
def find_mod_from_import_path_locally(
    import_path: ImportPath,
    file_extension: str,
    project_root: Path | None,
    *,
    should_exist: Literal[True],
) -> PathExists: ...
@overload
def find_mod_from_import_path_locally(
    import_path: ImportPath,
    file_extension: str,
    project_root: Path | None,
    *,
    should_exist: Literal[False],
) -> Path: ...


def find_mod_from_import_path_locally(
    import_path: ImportPath,
    file_extension: str = ".py",
    project_root: Path | None = None,
    *,
    should_exist: bool = True,
) -> PathExists | Path:
    """
    Given a module which import path is given by `import_path`,
    returns the module location as expected when iterating from the local working dir.
    """
    project_root = project_root or Path(".")
    path = project_root / convert_to_path(import_path, file_extension=file_extension)
    if not should_exist:
        return path
    if not path_exists(path):
        raise SmeltConfigError(
            f"Module {import_path} is not visible from current working dir:"
            f" {path} does not exists"
        )

    return path


@overload
def toggle_mod_path(path: str, to_type: Literal[ModpathType.IMPORT]) -> ImportPath: ...


@overload
def toggle_mod_path(path: str, to_type: Literal[ModpathType.FS]) -> FsPath: ...


def toggle_mod_path(
    path: str,
    to_type: ModpathType,
    src_folder: Path | None = None,
    exclude_root: bool = False,
) -> ImportPath | FsPath:
    """
    Changes the path to a module between 'import' style paths
    (a.b.c) and filesystem type (a/b/b).
    """
    src_folder = src_folder or Path(".")
    if get_modpath_type(path) == to_type:
        return cast(ImportPath | FsPath, path)

    match to_type:
        case ModpathType.IMPORT:
            path_obj = Path(path)
            resolved_path_obj = path_obj.resolve()
            resolved_src = src_folder.resolve()
            if not resolved_path_obj.is_relative_to(resolved_src):
                raise RuntimeError(f"`{src_folder}` is not a subpath of `path`")
            if exclude_root:
                resolved_path_obj = resolved_path_obj.parent
            rel_path = resolved_path_obj.relative_to(src_folder)
            modfile = rel_path.stem
            rel_path = rel_path.parent
            return cast(ImportPath, ".".join([*rel_path.parts, modfile]))

        case ModpathType.FS:
            path_chunks: list[str] = []
            package_or_namespace, *modules = path.split(".")
            if not modules:
                raise ValueError(
                    f"Modules import path only specifies the stem: {path}."
                    "You need to specify the whole import path"
                )
            if not exclude_root:
                path_chunks.append(package_or_namespace)
            *submodules, module = modules
            path_chunks.extend(submodules)
            module += ".py"
            path_chunks.append(module)
            final_path = src_folder.joinpath(*path_chunks)
            return cast(FsPath, str(final_path))

        case _ as unreachable:
            assert_never(unreachable)


def locate_module(
    mod_path: str,
    strategy: ModpathType = ModpathType.IMPORT,
    package_root: Path = Path("."),
) -> str:
    """
    Returns the full path to module at `mod_path`.
    If `mod_path` is passed a relative filepath, returns the absolute path.
    If it is passed as an import path ('.' seperated), tries importing and
    locates from the import"""
    match strategy:
        case ModpathType.IMPORT:
            import_path = locate_module_by_import_path(
                toggle_mod_path(mod_path, strategy)
            )
            if ctx := get_context():
                ctx.add_trace(
                    PathResolutionTrace(mod_path, mod_path, ModpathType.IMPORT)
                )
            return import_path

        case ModpathType.FS:
            fs_path = find_module_in_layout(
                toggle_mod_path(mod_path, strategy), package_root=package_root
            )
            if ctx := get_context():
                ctx.add_trace(PathResolutionTrace(fs_path, mod_path, ModpathType.FS))
            return fs_path

        case _ as unreachable:
            assert_never(unreachable)


def find_module_in_layout(mod_path: FsPath, package_root: str | None = None) -> str:
    """
    Given a module path on the filesystem,
    tries resolving that location within the package layout.
    """
    root = Path(package_root) if package_root else Path.cwd()
    src_location = root / "src"
    if src_location.exists():
        mod_potential_location = src_location / mod_path
        if mod_potential_location.exists():
            return str(mod_potential_location)

    mod_potential_location = root / mod_path
    if mod_potential_location.exists():
        return str(mod_potential_location)

    raise FileNotFoundError(f"Failed to locate {mod_path}")


@dataclass
class GenericExtension:
    """
    A wrapper on top of setuptools Extensions used by smelt.
    Used to provide metadata on how to manage these get_extension_suffixns in the install
    process, which is agnostic from the utility producing the extension in the first place.

    `import_path` keeps track of where the module should be placed in the final artifact.
    `src_path` keeps track of the original source files.
    `runtime` is optional and can be used to contain runtime utility module that's can
    be shared across modules (mypyc / Probably Nuitka in the future).
    """

    import_path: ImportPath
    src_path: PathExists
    dest_folder: Path
    extension: Extension
    runtime: Extension | None = None

    @classmethod
    def factory(
        cls,
        src_path: PathExists,
        import_path: ImportPath,
        extension: Extension | None = None,
        dest_folder: Path | None = None,
        runtime: Extension | None = None,
    ) -> Self:
        """
        Provides sound defaults for omitted fields.
        """
        extension = extension or Extension(
            name=get_module_name(import_path),
            sources=[src_path],
        )
        return cls(
            import_path=import_path,
            src_path=src_path,
            extension=extension,
            dest_folder=dest_folder or src_path.parent,
            runtime=runtime,
        )

    def get_dest_path(self, target_triple: str | None = None) -> Path:
        """
        Returns
        -------
        Path
            Full path for the final compiled .so file.
        """
        suffix = (
            sysconfig.get_config_var("EXT_SUFFIX")
            if target_triple is None
            else get_extension_suffix(target_triple)
        )
        ext_so_name = f"{self.name}{suffix}"
        return self.dest_folder / ext_so_name

    @property
    def name(self) -> ModuleName:
        """
        Returns the module name that given to the extension.
        """
        return get_module_name(self.import_path)

    def get_runtime_dest_path(self, target_triple: str | None = None) -> Path:
        """
        Returns
        -------
        Path
            Full path for the final runtime .so file.
        """
        suffix = (
            sysconfig.get_config_var("EXT_SUFFIX")
            if target_triple is None
            else get_extension_suffix(target_triple)
        )
        ext_so_name = f"{self.name}__mypyc{suffix}"
        return self.dest_folder / ext_so_name


class PackageRootPath(NamedTuple):
    module_import_path: ImportPath
    path: PathExists

    @classmethod
    def from_path(cls, path: PathExists) -> Self:
        package_name = path.parts[-1]
        if not is_valid_import_path(package_name):
            raise SmeltConfigError(
                f"Module path uses invalid characters: {package_name}"
            )
        return cls(package_name, path)


@dataclass
class PathSolver:
    known_roots: list[PackageRootPath] = field(default_factory=list)
    project_root: Path = Path(".")

    @classmethod
    def from_installed_import_paths(cls, *import_path: ImportPath):
        """
        Creates a PathSolver for the given import_paths, given that the reference modules
        are installed in the local environement.
        This relies on side-effects (modules location relies on specs given by importlib)
            and not suitable for build-time usage,
            but can be used as a shortcut for usage within an already installed environement
            (i.e. frontend usage).
        """
        roots = [
            PackageRootPath(p, locate_module_by_import_path(p)) for p in import_path
        ]
        return cls(roots)

    @overload
    def resolve_import_path(
        self,
        import_path: ImportPath,
        file_extension: str,
        *,
        should_exist: Literal[True],
    ) -> PathExists: ...
    @overload
    def resolve_import_path(
        self,
        import_path: ImportPath,
        file_extension: str,
        *,
        should_exist: Literal[False],
    ) -> Path: ...

    def resolve_import_path(
        self,
        import_path: ImportPath,
        file_extension: str = ".py",
        *,
        should_exist: bool = True,
    ) -> Path | PathExists:
        # TODO: detect when target module is already native
        for root_path, path in self.known_roots:
            if import_path == root_path:
                return path
            if not import_path.startswith(root_path):
                continue
            root = Path(path)
            sub_import_path = import_path.replace(root_path + ".", "")
            assert is_valid_import_path(sub_import_path), (
                f"internal path substitution logic error {sub_import_path}"
            )

            subpath = convert_to_path(sub_import_path)
            module_path = root / subpath
            if should_exist and not path_exists(module_path):
                raise SmeltConfigError(
                    f"Module {import_path} should resolve to this location "
                    f"but we did not find it: {module_path}"
                )
            return module_path

        return find_mod_from_import_path_locally(
            import_path,
            file_extension=file_extension,
            project_root=self.project_root,
            should_exist=should_exist,
        )



"""
Unit tests for the context management.

@date: 26.11.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

from typing import Generator

import pytest

from smelt.context import (
    clear_contexts,
    create_context_if_enabled,
    enable_global_context,
    get_context,
    is_global_context_enabled,
    reset_contexts,
)
from smelt.utils import ModpathType, PathResolutionTrace


@pytest.fixture(autouse=True)
def clear_contexts_before_and_after() -> Generator[None, None, None]:
    clear_contexts()
    try:
        yield
    finally:
        clear_contexts()


def test_global_context_toggling() -> None:
    assert is_global_context_enabled() is False
    enable_global_context()
    assert is_global_context_enabled() is True
    reset_contexts()
    assert is_global_context_enabled() is True
    clear_contexts()
    assert is_global_context_enabled() is False


def test_get_global_context() -> None:
    assert get_context() is None
    enable_global_context()
    ctx1 = get_context()

    ctx2 = get_context()
    assert ctx1 is ctx2
    clear_contexts()
    assert get_context() is None


def test_add_trace() -> None:
    enable_global_context()
    ctx = get_context()
    import_path = "a.b"
    fs_path = "a/b"
    trace1 = PathResolutionTrace(
        import_path, fs_path, resolution_type=ModpathType.IMPORT
    )
    trace2 = PathResolutionTrace(import_path, fs_path, resolution_type=ModpathType.FS)
    assert ctx is not None
    ctx.add_trace(trace1)
    ctx.add_trace(trace2)
    # TODO: add snapshot. For now only checking basic stuff
    rendered = ctx.render()
    assert rendered.count("a/b") == 2
    assert rendered.count("a.b") == 2


class DummyContext:
    def render(self) -> str:
        return "dummy"


def test_get_context() -> None:
    enable_global_context()
    ctx1 = create_context_if_enabled("ctx1", DummyContext)
    ctx2 = create_context_if_enabled("ctx2", DummyContext)
    assert isinstance(ctx1, DummyContext)
    assert isinstance(ctx2, DummyContext)
    assert ctx1 is not ctx2
    assert get_context("ctx1") is ctx1
    assert get_context("ctx2") is ctx2
    global_ctx = get_context()
    assert global_ctx is not None
    assert get_context() is global_ctx
    rendered = global_ctx.render()
    assert rendered.count("dummy") == 2



"""
Command-line interface for Smelt

@date: 12.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import logging
import os
import shutil
import sys
import tomllib
import warnings
from contextlib import contextmanager, chdir
from pathlib import Path
from typing import Callable, Generator, Literal, NoReturn, ParamSpec, TypeVar

import click
from click import Context, Parameter, ParamType

from smelt.backend import (
    compile_cython_extensions,
    compile_mypyc_extensions,
    nuitkaify_module,
    run_backend,
)
from smelt.compiler import SupportedPlatforms, compile_extension
from smelt.config import (
    CythonExtension,
    MypycModule,
    NuitkaModule,
    SmeltConfig,
    TomlData,
    auto_detect_is_build_hook,
    toml_get_nested_section,
)
from smelt.context import enable_global_context, get_context
from smelt.utils import (
    ImportPath,
    PathExists,
    PathSolver,
    SmeltError,
    is_valid_import_path,
    path_exists,
)


class SmeltConfigError(SmeltError): ...


P = ParamSpec("P")
R = TypeVar("R")

SMELT_ASCCI_ART: str = r"""
 ____                 _ _
/ ___| _ __ ___   ___| | |_
\___ \| '_ ` _ \ / _ \ | __|
 ___) | | | | | |  __/ | |_
|____/|_| |_| |_|\___|_|\__|

"""

add_logging_option = click.option(
    "-l",
    "--logging-level",
    type=click.Choice(list(logging._nameToLevel), case_sensitive=False),
    help="Logging level to apply. Logs are emitted to stdout",
    default="warning",
)


class CliImportPath(ParamType):
    """
    A tiny wrapper for click to verify import paths validity automatically.
    """

    name = "import_path"

    def convert(
        self, value: str, param: Parameter | None, ctx: Context | None
    ) -> ImportPath:
        _ = param
        _ = ctx
        if not is_valid_import_path(value):
            self.fail(f"{value} is not a valid Python import path")
        return value


class CliExistingPath(ParamType):
    """
    A tiny wrapper for click to verify import paths validity automatically.
    """

    name = "existing_path"

    def convert(
        self, value: str, param: Parameter | None, ctx: Context | None
    ) -> PathExists:
        _ = param
        _ = ctx
        path = Path(value)
        if not path_exists(path):
            self.fail(f"{value} not found")
        return path


def wrap_smelt_errors(
    should_exist: bool = True, exit_code: int = 1
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Captures `SmeltError` exceptions and displays them to the user in a nicer way.
    """

    @contextmanager
    def wrapper() -> Generator[None, None, None]:
        try:
            yield
        except SmeltError as exc:
            click.echo("/!\\  [Smelt] An error occured:")
            click.echo(exc)
            if should_exist:
                sys.exit(exit_code)

    return wrapper()


def parse_config_from_pyproject(
    toml_data: TomlData,
    is_configured_as_build_hook: bool | None = None,
    project_root: Path | None = None,
) -> SmeltConfig:
    """
    Extracts Smelt config from TOML data coming out of a pyproject.toml
    """
    is_configured_as_build_hook = (
        is_configured_as_build_hook
        if is_configured_as_build_hook is not None
        else auto_detect_is_build_hook(toml_data)
    )
    tool_config = toml_data.get("tool", {})
    config_path = (
        ("tool", "hatch", "build", "hooks", "smelt")
        if is_configured_as_build_hook
        else ("tool", "smelt")
    )
    if not isinstance(tool_config, dict):
        raise SmeltConfigError(
            f"`tool` section in toml data is not a dictionary, got {tool_config}. "
            "Does the TOML data come from a valid pyproject ?"
        )
    smelt_config = toml_get_nested_section(toml_data, *config_path)
    if smelt_config is None:
        raise SmeltConfigError("No smelt config defined in pyproject")

    if not isinstance(smelt_config, dict):
        raise SmeltConfigError(
            f"`smelt` section should be a dictionary, got {smelt_config}. "
        )
    return SmeltConfig.from_toml_data(smelt_config, project_root=project_root)


def error_exit(msg: str, code: int = 1) -> NoReturn:
    """
    A helper exiting the program with the given `msg` on error.
    """
    click.echo(msg)
    sys.exit(code)


@click.group()
def smelt() -> None:
    """
    Entrypoint for Smelt frontend
    """
    enable_global_context()
    click.echo(SMELT_ASCCI_ART)


@smelt.command()
@click.option(
    "-p",
    "--path",
    default=".",
    type=CliExistingPath(),
)
@wrap_smelt_errors()
def show_config(*, path: PathExists) -> None:
    """
    Shows the smelt config as defined in the passed file
    """
    from pprint import pprint

    try:
        with (path / "pyproject.toml").open("rb") as f:
            toml_data = tomllib.load(f)
    except FileNotFoundError:
        click.echo("No pyproject.toml not found.")
        return
    pprint(parse_config_from_pyproject(toml_data, project_root=path))


@smelt.command()
@click.option(
    "-p",
    "--package-path",
    default=".",
    type=CliExistingPath(),
)
@add_logging_option
@click.option(
    "-r", "--report", type=str, default=None, help="Produces a report at the given path"
)
@wrap_smelt_errors()
def build_standalone_binary(
    package_path: PathExists, logging_level: str, report: str | None
) -> None:
    levelno = logging._nameToLevel[logging_level]
    logging.basicConfig(level=levelno)
    try:
        with (package_path / "pyproject.toml").open("rb") as f:
            toml_data = tomllib.load(f)
    except FileNotFoundError:
        click.echo("No pyproject.toml not found.")
        return
    config = parse_config_from_pyproject(toml_data, project_root=package_path)
    config.load_env()
    path_solver = config.get_path_solver(project_root=package_path)
    try:
        run_backend(config, stdout="stdout", path_solver=path_solver)
    except Exception as e:
        click.echo(f"Error during build: {e}")
    if report is not None:
        global_context = get_context()
        assert global_context is not None
        Path(report).write_text(global_context.render())


@smelt.command()
@click.argument(
    "entrypoint-path",
    type=ImportPath,
)
@add_logging_option
@wrap_smelt_errors()
def nuitkaify(entrypoint_path: ImportPath, logging_level: str) -> None:
    """
    Standalone command to run the nuitka wrapper in this package.
    This is mainly intended for manual self-testing, if you only need nuitka
    features you should probably just call nuitka directly.
    """
    from smelt.nuitkaify import compile_with_nuitka

    levelno = logging._nameToLevel[logging_level]
    logging.basicConfig(level=levelno)
    compile_with_nuitka(entrypoint_path, stdout="stdout")


@smelt.command()
@click.argument(
    "module-import-path",
    type=CliImportPath(),
)
@click.option(
    "-b",
    "--backend",
    default="nuitka",
    type=click.Choice(["mypyc", "nuitka", "cython"]),
    help="How to compile the module",
)
@click.option(
    "-cp",
    "--crosscompile",
    type=click.Choice([platform.value for platform in SupportedPlatforms]),
    default=None,
)
@wrap_smelt_errors()
def compile_module(
    module_import_path: ImportPath,
    backend: Literal["mypyc", "nuitka", "cython"],
    crosscompile: str | None,
) -> None:
    """
    Standalone command to run the nuitka wrapper in this package.
    This is mainly intended for manual self-testing, if you only need nuitka
    features you should probably just call nuitka directly.
    """
    path_solver = PathSolver.from_installed_import_paths(module_import_path)
    click.echo(f"Compiling module {module_import_path}")
    try:
        module_source = path_solver.resolve_import_path(
            module_import_path, should_exist=True
        )
    except SmeltConfigError as exc:
        error_exit(str(exc))

    if backend == "nuitka":
        config = NuitkaModule(module_import_path, module_source)
        generic_ext = nuitkaify_module(config, path_solver, stdout="stdout")

    elif backend == "mypyc":
        target_platform = SupportedPlatforms(crosscompile) if crosscompile else None
        target_triple_name = (
            None if target_platform is None else target_platform.get_triple_name()
        )
        modules = [MypycModule(module_import_path)]
        (generic_ext,) = compile_mypyc_extensions(modules, path_solver)

    elif backend == "cython":
        modules = [CythonExtension(module_import_path)]
        (generic_ext,) = compile_cython_extensions(modules, path_solver=path_solver)
    compiled_so = compile_extension(generic_ext.extension)
    dest_path = generic_ext.dest_folder / compiled_so
    shutil.move(compiled_so, dest_path)
    if runtime := generic_ext.runtime:
        runtime_compiled_so = compile_extension(runtime)
        shutil.move(runtime_compiled_so, generic_ext.dest_folder / runtime_compiled_so)
    click.echo(f"Compiled so path: {dest_path}")


@smelt.command
@click.option(
    "-p",
    "--package",
    type=CliExistingPath(),
    help="Path the the package to build extensions for, expects to find a pyproject.toml",
    default=Path.cwd(),
)
def build_extensions(*, package: PathExists) -> None:
    """
    Runs the smelt backend on the passed project and builds all extensions
    defined by smelt.
    """
    pyproject_path = package / "pyproject.toml"
    if not path_exists(pyproject_path):
        error_exit("No pyproject.toml found in passed folder")
    with pyproject_path.open("rb") as f:
        try:
            toml_data = tomllib.load(f)
        except tomllib.TOMLDecodeError as exc:
            error_exit(f"Invalid TOML file [{pyproject_path}]: {exc}")
        config = parse_config_from_pyproject(toml_data, project_root=package)
        path_solver = config.get_path_solver(project_root=package)
        run_backend(
            config, stdout="stdout", path_solver=path_solver, without_entrypoint=True
        )



"""
Build backend implementation for smelt.

@date: 12.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import logging
import os
import shutil
import warnings
from pathlib import Path
from typing import Any, Iterable


from smelt.compiler import compile_extension, compile_zig_module

from mypyc.build import mypycify
from smelt.nuitkaify import Stdout, compile_with_nuitka, nuitkaify_module
from smelt.utils import (
    GenericExtension,
    ModpathType,
    locate_module,
    PathSolver,
)
from smelt.config import SmeltConfig, MypycModule, CythonExtension

# TODO: replace .so references to a variable that's set to .so
# for Unix-like and .dll for Windows

_logger = logging.getLogger(__name__)


def compile_mypyc_extensions(
    modules: Iterable[MypycModule],
    path_solver: PathSolver | None = None,
) -> list[GenericExtension]:
    """
    Compiles all mypy extensions defined in `mypyc_config` for the project found at `project_root`
    """
    path_solver = path_solver or PathSolver()
    built_extensions: list[GenericExtension] = []
    for module in modules:
        module_import_path = module.import_path
        ext_path = module.source or path_solver.resolve_import_path(module_import_path)
        runtime, module_ext = mypycify([str(ext_path)], include_runtime_files=True)
        mypyc_ext = GenericExtension.factory(
            src_path=ext_path,
            import_path=module_import_path,
            extension=module_ext,
            runtime=runtime,
            dest_folder=ext_path.parent,
        )

        module_so_path = compile_extension(mypyc_ext.extension)
        runtime_so_path = compile_extension(mypyc_ext.runtime)
        so_dest_path = str(mypyc_ext.get_dest_path())
        runtime_dest_path = str(mypyc_ext.get_runtime_dest_path())
        shutil.move(runtime_so_path, runtime_dest_path)
        shutil.move(module_so_path, so_dest_path)
        _logger.info("Built extensions %s @ %s", module_import_path, so_dest_path)
        if mypyc_ext.runtime:
            _logger.info("-> %s runtime: %s", module_import_path, runtime_dest_path)
        built_extensions.append(mypyc_ext)
    return built_extensions


def compile_cython_extensions(
    modules: list[CythonExtension],
    options: dict[str, Any] | None = None,
    path_solver: PathSolver | None = None,
) -> list[GenericExtension]:
    """
    Compiles all the cython extensions as defined in `cython_config`
    """
    path_solver = path_solver or PathSolver()
    options = options or {}
    try:
        from Cython.Build import cythonize
    except ImportError as exc:
        raise ImportError(
            "Cython is not installed. consider installing smelt with [cython] extra"
        ) from exc
    extensions: list[GenericExtension] = []

    for module in modules:
        import_path = module.import_path
        source_path = str(module.source or path_solver.resolve_import_path(import_path))
        cython_ext = cythonize(source_path, **options)
        assert len(cython_ext) == 1, (
            "Passed on source file to cython yet it produced more than one extension"
        )
        (base_ext,) = cython_ext
        ext_name = import_path.split(".")[-1]
        base_ext.name = ext_name
        generic_ext = GenericExtension.factory(
            src_path=source_path,
            import_path=import_path,
            extension=base_ext,
            dest_folder=Path(source_path).parent,
        )
        extensions.append(generic_ext)
    return extensions


def run_backend(
    config: SmeltConfig,
    stdout: Stdout | None = None,
    path_solver: PathSolver | None = None,
    strategy: ModpathType = ModpathType.FS,
    *,
    without_entrypoint: bool = False,
) -> None:
    """
    Runs the whole backend pipeline:
    * C extensions compilation
    * mypyc extensions
    * Nuitka compilation
    """
    path_solver = path_solver or config.get_path_solver()
    # Starting with C extensions
    warnings.warn(
        "`run_backend` implementation is not fully implemented yet and will only "
        "compile C extensions"
    )
    for zig_mod in config.zig_modules:
        compile_zig_module(
            zig_mod.name,
            zig_mod.folder,
            zig_mod.import_path,
            flags=zig_mod.flags,
            path_solver=path_solver,
        )

    for native_extension in config.c_extensions:
        sources = native_extension.sources
        if len(sources) > 1:
            raise NotImplementedError("Not supported yet")
        c_extension_path = sources[0]
        parent_folder_path = Path(c_extension_path).parent
        # TODO: we should probably run that logic in temp folder
        built_so_path = compile_extension(c_extension_path)
        so_final_path = parent_folder_path / os.path.basename(built_so_path)
        shutil.move(built_so_path, so_final_path)

    # Note: mypyc has a runtime shipped as a separate extension
    # this runtime should be named modname__mypy
    # we need to keep track of it to include to nuitka,
    # as it would be invisible otherwise
    shared_runtime_extensions: set[str] = set()
    collected_extensions: list[GenericExtension] = []
    built_mypyc_extensions = compile_mypyc_extensions(config.mypyc_modules, path_solver)
    for ext in built_mypyc_extensions:
        if ext.runtime:
            shared_runtime_extensions.add(ext.runtime.name)
    # cython extensions
    collected_extensions.extend(
        compile_cython_extensions(config.cython_modules, path_solver=path_solver)
    )
    for nuitka_mod in config.nuitka_modules:
        collected_extensions.append(
            nuitkaify_module(nuitka_mod, path_solver=path_solver)
        )

    for generic_ext in collected_extensions:
        module_so_path = compile_extension(generic_ext.extension)
        shutil.move(module_so_path, str(generic_ext.get_dest_path()))
        if generic_ext.runtime:
            runtime_so_path = compile_extension(generic_ext.extension)
            shutil.move(runtime_so_path, str(generic_ext.get_runtime_dest_path()))
    # nuitka entrypoint compilation
    without_entrypoint = without_entrypoint or config.entrypoint is None
    if not without_entrypoint:
        entrypoint_file = locate_module(
            config.entrypoint, strategy=strategy, package_root=path_solver
        )
        compile_with_nuitka(
            entrypoint_file, stdout=stdout, include_modules=shared_runtime_extensions
        )



"""
Config definitions for Smelt.

@date: 19.02.2026
@author: Baptiste Pestourie
"""

from __future__ import annotations

from pathlib import Path
import os
from dataclasses import MISSING, dataclass, field, fields, asdict
from re import M
from typing import TYPE_CHECKING, Any, Iterable, Self

from smelt.utils import (
    ImportPath,
    PackageRootPath,
    PathExists,
    PathSolver,
    SmeltConfigError,
    SmeltError,
    assert_is_valid_import_path,
    assert_path_exists,
    is_valid_import_path,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

_CONVERSIONS = {
    "ImportPath": assert_is_valid_import_path,
    "PathExists": assert_path_exists,
    "list[PathExists]": lambda values: [assert_path_exists(p) for p in values],
}


def assert_type_is(obj: object, typ: type) -> None:
    if not isinstance(obj, typ):
        raise SmeltConfigError(f"Expected type {typ}, got {type(obj)}")


def convert_path(
    path_decl: str | list[str], type_hint: str, project_root: Path | None
) -> ImportPath | PathExists | Path | list[PathExists]:
    if type_hint == "ImportPath":
        assert_type_is(path_decl, str)
        return assert_is_valid_import_path(path_decl)

    match type_hint:
        case "PathExists" | "PathExists | None":
            assert_type_is(path_decl, str)
            path = (
                project_root / path_decl
                if project_root is not None
                else Path(path_decl)
            )

            assert_path_exists(path)
            return path
        case "list[PathExists]":
            assert_type_is(path_decl, list)
            return [
                assert_path_exists(project_root / p if project_root else Path(p))
                for p in path_decl
            ]

        case "list[str]":
            assert_type_is(path_decl, list)
            return path_decl

        case "str":
            assert_type_is(path_decl, str)
            return path_decl

        case _:
            assert_type_is(path_decl, str)
            raise NotImplementedError(f"Unsupported type hint: {type_hint}")


type ConfigContext = Iterable[str]


def _format_context(context: ConfigContext) -> str:
    """
    A small helper that builds a human-friendly hint from
    the path of a parameter in the config.
    Meant to be used when reporting config errors.
    """
    if not context:
        return ""

    *nodes, leaf = context

    param_path = ".".join(nodes) + ":" + leaf if nodes else leaf
    return f"In [{param_path}]:"


def build_datacls_from_toml[T: DataclassInstance](
    datacls: type[T],
    toml_data: TomlData,
    context: ConfigContext | None = None,
    project_root: Path | None = None,
) -> T:
    context = context if context is not None else []
    sentinel = object()
    kwargs: dict[str, object] = {}
    for f in fields(datacls):
        local_ctx = list(context)
        field_name = f.name
        value_decl = toml_data.get(field_name, sentinel)
        if value_decl is sentinel:
            if f.default is MISSING and f.default_factory is None:
                raise SmeltConfigError(
                    f"{_format_context(local_ctx)}Missing mandatory argument: {f.name}"
                )
            continue
        local_ctx.append(f.name)
        assert isinstance(f.type, str), (
            "Expected annotations from __future__ to be used"
        )
        try:
            value = convert_path(value_decl, f.type, project_root)
        except SmeltError as exc:
            raise SmeltConfigError(f"{_format_context(local_ctx)}{exc}") from exc
        kwargs[field_name] = value
    return datacls(**kwargs)


type _TomlData = str | list[_TomlData] | dict[str, _TomlData]
type TomlData = dict[str, _TomlData]


def auto_detect_is_build_hook(toml_data: TomlData) -> bool:
    """
    Given the extracted TOML config `toml_data`,
    detects whether Smelt Config was passed as a build hook
    (in which case it would be nested under whatever subsection_name
    the build backend uses for hooks), or a tool config.
    """
    has_tool_config = "smelt" in toml_data.get("tool", {})
    has_build_hook_conf = "smelt" in toml_get_nested_section(
        toml_data, "tool", "hatch", "build", "hooks"
    )
    if has_tool_config and has_build_hook_conf:
        # TODO: for now, not allowing this.
        # We can however consider using the hatch one only for build time
        # and the tool one for CLI use.
        # that can get confusing though.
        raise SmeltConfigError(
            "Smelt configuration found both in [tool.smelt] and "
            "[tool.hatch.build.hooks.smelt]. Please keep only one."
        )
    if has_build_hook_conf:
        return True
    if has_tool_config:
        return False
    raise SmeltConfigError("No smelt config detected")


def toml_get_nested_section(toml_data: TomlData, *path: str) -> _TomlData:
    """
    Extracts the sub section given by `path` from `toml_data`.
    Verifies that the extracted TOML object is a dictionary.

    Raises
    ------
    SmeltConfigError
        If the section is not found or if the found object is not a section.
    """
    ctx: list[str] = []
    section: _TomlData = toml_data
    for subsection_name in path:
        ctx.append(subsection_name)
        section = section.get(subsection_name, {})
        if not isinstance(section, dict):
            raise SmeltConfigError(
                f"{_format_context(ctx)}Expected section, found {section}"
            )
    return section


@dataclass
class NuitkaModule:
    import_path: ImportPath
    source: PathExists | None = None


@dataclass
class NativeExtension:
    import_path: ImportPath
    sources: list[PathExists]


@dataclass
class CythonExtension:
    import_path: ImportPath
    source: PathExists | None = None


@dataclass
class MypycModule:
    import_path: ImportPath
    source: PathExists | None = None


@dataclass
class ZigModule:
    name: str
    import_path: ImportPath
    folder: PathExists = assert_path_exists(".")
    flags: list[str] = field(default_factory=list)


@dataclass
class SmeltConfig:
    """
    Defines how the smelt backend should run
    """

    packages_location: dict[str, str] = field(default_factory=dict)
    mypyc_options: dict[str, Any] = field(default_factory=dict)
    mypyc_modules: list[MypycModule] = field(default_factory=list)
    cython_options: dict[str, Any] = field(default_factory=dict)
    cython_modules: list[CythonExtension] = field(default_factory=list)
    nuitka_modules: list[NuitkaModule] = field(default_factory=list)
    c_extensions: list[NativeExtension] = field(default_factory=list)
    zig_modules: list[ZigModule] = field(default_factory=list)
    entrypoint: str | None = None
    debug: bool = False

    @classmethod
    def from_toml_data(
        cls, toml_data: dict[str, Any], project_root: Path | None = None
    ) -> Self:
        # native code
        native_extensions_decl = toml_data.pop("c_extensions", [])
        native_extensions = [
            build_datacls_from_toml(NativeExtension, decl, project_root=project_root)
            for decl in native_extensions_decl
        ]
        # zig modules
        zig_modules_decl = toml_data.pop("zig_modules", [])
        zig_modules = [
            build_datacls_from_toml(ZigModule, decl, project_root=project_root)
            for decl in zig_modules_decl
        ]
        # mypyc modules
        mypyc_modules_decl = toml_data.pop("mypyc_modules", [])
        mypyc_modules = [
            build_datacls_from_toml(MypycModule, decl, project_root=project_root)
            for decl in mypyc_modules_decl
        ]

        # cython
        cython_modules_decl = toml_data.pop("cython_modules", [])
        cython_modules = [
            build_datacls_from_toml(CythonExtension, decl, project_root=project_root)
            for decl in cython_modules_decl
        ]
        # nuitka
        nuitka_modules_decl = toml_data.pop("nuitka_modules", [])
        nuitka_modules = [
            build_datacls_from_toml(NuitkaModule, decl, project_root=project_root)
            for decl in nuitka_modules_decl
        ]

        return cls(
            mypyc_modules=mypyc_modules,
            c_extensions=native_extensions,
            zig_modules=zig_modules,
            cython_modules=cython_modules,
            nuitka_modules=nuitka_modules,
            **toml_data,
        )

    def get_path_solver(self, project_root: Path | None = None) -> PathSolver:
        """
        Builds a PathSolver based on the package configuration.
        """
        root = project_root or Path.cwd()
        return PathSolver(
            known_roots=[
                PackageRootPath(alias, assert_path_exists(root / path))
                for alias, path in self.packages_location.items()
            ],
            project_root=root,
        )

    def __str__(self) -> str:
        """
        A human-friendly stringified version of this config.
        """
        lines: list[str] = []
        for field_name, value in asdict(self).items():
            if isinstance(value, list):
                value = ",".join(value)
            if isinstance(value, dict):
                value = "".join(
                    ("\n * " + f"{key} -> {val}" for key, val in value.items())
                )
            lines.append(f"{field_name:20}: {value}")
        return "\n".join(lines)

    def load_env(self) -> None:
        """
        Updates internal values based on set environement variables.
        """
        if os.environ.get("SMELT_DEBUG"):
            self.debug = True



"""
Main configuration for pytest.
Tests in this folder are inteded to be run with nox,
but running manually with pytest is supported for certain
specific cases.

@date: 11.12.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations
from pytest import Parser


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--package",
        action="store",
        default=None,
        help="Package name from which to test fibonacci function",
    )



"""
Test suite for smelt frontend.

@date: 13.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import pytest
import tomllib

from smelt.frontend import TomlData, parse_config_from_pyproject

SAMPLE_CONFIG: str = """\
[tool.smelt]
entrypoint="minimal.cli"

[tool.smelt.c_extensions]
"minimal.hello"="src/minimal/hello.c"

[tool.smelt.mypyc]
"minimal.fib" = "src/minimal/fib.py"
"""


@pytest.fixture
def toml_data() -> TomlData:
    """
    Example TOML data as it would be extracted from pyproject.toml
    """
    return tomllib.loads(SAMPLE_CONFIG)


def test_parse(toml_data: TomlData) -> None:
    """
    Verifies that the pyproject parse extracts Smelt config properly
    """
    config = parse_config_from_pyproject(toml_data)
    assert config.entrypoint == "minimal.cli"
    assert config.c_extensions == {"minimal.hello": "src/minimal/hello.c"}
    assert config.mypyc == {"minimal.fib": "src/minimal/fib.py"}



"""
Build hook hatchling backend.

@date: 03.09.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

from functools import cached_property
from dataclasses import fields

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl

from smelt.config import SmeltConfig
from smelt.backend import run_backend
from smelt.utils import ModpathType


class HatchlingBuildHook(BuildHookInterface):
    PLUGIN_NAME = "smelt"

    @cached_property
    def is_debug(self) -> bool:
        if not self.smelt_config.debug:
            return False
        print("Smelt: SMELT_DEBUG is set, enabling debug mode")
        return True

    @cached_property
    def smelt_config(self) -> SmeltConfig:
        try:
            config = SmeltConfig.from_toml_data(self.config)
        except Exception as exc:
            raise ValueError(
                "Smelt config is invalid:"
                f"Current config: {self.config}"
                "Valid parameters are:\n"
                f"{[f.name for f in fields(SmeltConfig)]}"
            ) from exc
        config.load_env()
        return config

    def debug_log(self, message: str) -> None:
        """
        Prints `message` if debug mode is set.
        """
        if not self.is_debug:
            return
        print(message)

    def initialize(self, version: str, build_data: dict[str, object]) -> None:
        if self.target_name == "sdist":
            # disabling ourselves - we only want to include source code.
            # TODO: consider adding an environment variable to force building
            # extensions even in sdist mode, if there's ever a use for that
            return

        self.debug_log(f"Smelt: Calling build hook with config:\n{self.config}")
        try:
            run_backend(
                self.smelt_config, strategy=ModpathType.FS, without_entrypoint=True
            )
        except Exception as exc:
            raise RuntimeError(f"Smelt build failed: {exc}")


@hookimpl
def hatch_register_build_hook() -> type[BuildHookInterface]:
    """
    Registers Smelt's build hook as a hatch plugin
    """
    return HatchlingBuildHook



"""
Isolated functional tests, based on Nox.
Spawns packages with various config and verifies
that the Smelt build hook processes them properly.

@date: 11.12.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import os
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator

import nox

if TYPE_CHECKING:
    from nox import Session


@dataclass
class PackageConfig:
    """
    Defines which type of extensions should be generated
    in the temporary package.
    """

    has_nuitka: bool = True
    has_mypyc: bool = True
    has_cython_pyx: bool = True
    has_cython_py: bool = True

    def build_pyproject(self, project_name: str) -> str:
        """
        Generates the `pyproject.toml` content based
        on the defined extensions types.
        """
        sections: list[str] = [PYPROJECT_TEMPLATE]
        if self.has_nuitka:
            sections.append(SMELT_NUITKA_SECTION)
        if self.has_mypyc:
            sections.append(SMELT_MYPYC_SECTION)
        if self.has_cython_py:
            sections.append(SMELT_CYTHON_SECTION)

        return "\n".join([h.format(project_name=project_name) for h in sections])


FIB_CYTHON = """\
def fibx(int n):
    cdef int i, a, b
    a, b = 0, 1
    for i in range(n):
        a, b = a + b, a
    return a\
"""

FIB_PYTHON = """\
def {fib_name}(n: int) -> int:
    if n <= 1:
        return n
    else:
        return {fib_name}(n - 2) + {fib_name}(n - 1)\
"""

PYPROJECT_TEMPLATE = """\
[build-system]
requires = ["hatchling", "smelt", "mypy[mypyc]", "Cython", "Nuitka"]
build-backend = "hatchling.build"

[project]
name = "{project_name}"
version = "0.1.0"
dependencies = ["click"]

[project.scripts]
compute-fib = "{project_name}.cli:compute_fib"

[tool.hatch.build.targets.wheel]
packages = ["src/{project_name}"]
"""

SMELT_NUITKA_SECTION = """\
[tool.hatch.build.hooks.smelt]
entrypoint = "{project_name}.cli"
"""

SMELT_CYTHON_SECTION = """\
[[tool.hatch.build.hooks.smelt.cython_modules]]
import_path = "{project_name}.fib_cython"
source = "src/{project_name}/fib_cython.pyx"
"""

SMELT_MYPYC_SECTION = """\
[[tool.hatch.build.hooks.smelt.mypyc_modules]]
import_path="{project_name}.fib"
"""

CLI_ENTRYPOINT = """\
import click
from {project_name}.fib_pure_python import fib_pure_python
@click.command()
@click.argument("n", type=int)
def compute_fib(n: int) -> None:
    click.echo(fib_pure_python(n))
"""

NUITKA_ONLY_CONFIG = PackageConfig(
    has_nuitka=True,
    has_cython_py=False,
    has_cython_pyx=False,
    has_mypyc=False,
)


@contextmanager
def spawn_package(
    project_name: str, config: PackageConfig | None = None
) -> Generator[Path, None, None]:
    """
    Generates a package in a temporary folder,
    with the `pyproject.toml` and source files generated
    as defined by `config`.
    """
    config = config or PackageConfig()
    with tempfile.TemporaryDirectory() as dir:
        root = Path(dir)
        pyproject = root / "pyproject.toml"
        pyproject.write_text(config.build_pyproject(project_name=project_name))

        src_dir: Path = root / "src" / project_name
        os.makedirs(src_dir)

        (src_dir / "fib_pure_python.py").write_text(
            FIB_PYTHON.format(fib_name="fib_pure_python")
        )
        if config.has_mypyc:
            (src_dir / "fib_mypyc.py").write_text(
                FIB_PYTHON.format(fib_name="fib_mypyc")
            )
        if config.has_cython_py:
            (src_dir / "fib_cython.py").write_text(
                FIB_PYTHON.format(fib_name="fib_cython")
            )
        if config.has_cython_pyx:
            (src_dir / "fib_cython.pyx").write_text(
                FIB_CYTHON.format(fib_name="fib_cython_pyx")
            )
        if config.has_nuitka:
            (src_dir / "cli.py").write_text(
                CLI_ENTRYPOINT.format(project_name=project_name)
            )
        yield root


def test_fib_sanity_check() -> None:
    """
    Checks the validaity of the fibonacci code template.
    """
    namespace: dict[str, Any] = {}
    exec(FIB_PYTHON.format(fib_name="fib"), locals=namespace, globals=namespace)
    assert "fib" in namespace
    assert eval("fib(10)", locals=namespace, globals=namespace) == 55


@nox.session
def spawn_package_sanity_check(session: Session) -> None:
    """
    Checks that a basic package can be generated and installed.
    """
    project_name = "testproject"
    with spawn_package(
        project_name=project_name, config=NUITKA_ONLY_CONFIG
    ) as proj_dir:
        session.install(str(proj_dir))
        # sanity check: importing the package we jsut installed
        session.run("python", "-c", f'"import {project_name}"')
        session.install("pytest")
        # installing smelt itself
        # NOTE: relative location is hard-coded, if this tests folder gets moved around
        # this should be updated accordingly
        session.install("../..")
        session.run("python", "-m", "pytest", "--package", project_name)



"""
Test suite for nuitka operations

@date: 11.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import contextlib
import os
import subprocess
import tempfile

from test_compiler import get_module_path

from smelt.nuitkaify import compile_with_nuitka


def fib(n: int) -> int:
    if n <= 1:
        return n
    else:
        return fib(n - 2) + fib(n - 1)


def test_nuitkaify() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        # Set the temporary directory as the current working directory
        standalone_app = get_module_path("entrypoint")
        with contextlib.chdir(tmpdir):
            # Compile the standalone app using nuitka
            compile_with_nuitka(str(standalone_app), stdout="stdout")
        expected_exe_path = os.path.join(tmpdir, "entrypoint.bin")
        assert os.path.exists(
            expected_exe_path
        ), "Nuitka did not produce the expected executable."

        # testing the produced exe - computing fib(10), checking if we get it on stdout
        p = subprocess.run(
            [expected_exe_path, "10"], check=True, stdout=subprocess.PIPE
        )
        assert p.returncode == 0, "The produced executable did not run successfully."
        assert p.stdout.decode().rstrip() == str(fib(10))



"""
Wrapper on top of nuitka to compile a Python script into a standalone executable.

Currently nuitka is called as a subprocess, as it would be from `python -m nuitka`.
Options are passed as CLI arguments.

This is the simple option as nuitka is not really designed for library use: some of the business logic
is run on import, a few critical components are handled global variables, so there a some major drawbacks to
trying to import the code and call directly.
This might be changed later.

@date: 11.06.2025
@author: Baptiste Pestourie
"""

from __future__ import annotations

import logging
import os
from dataclasses import dataclass
import sys
from pathlib import Path
from typing import Final, Iterable, Iterator, Literal

from setuptools import Extension

from .process import call_command

from smelt.context import get_context
from smelt.utils import GenericExtension, PathSolver
from smelt.config import NuitkaModule

_logger = logging.getLogger(__name__)


NUITKA_ENTRYPOINT: Final[tuple[str, ...]] = (sys.executable, "-m", "nuitka")

type Stdout = Literal["stdout", "logger"]

# TODO: this should be built dynamically, obviously
NUITKA_MACROS = [
    ("_XOPEN_SOURCE", None),
    ("__NUITKA_NO_ASSERT__", None),
    ("_NUITKA_CONSTANTS_FROM_CODE", None),
    ("_NUITKA_FROZEN", 0),
    # TODO:
    # Note: seems that that one was NUITKA_MODULE_MODE
    # and was renamed around Nuitka 2.7.9 to _NUIKA_MODULE
    ("_NUITKA_MODULE_MODE", 1),
]

NUITKA_MINIMAL_FLAGS: Final[tuple[str, ...]] = (
    "-std=c11",
    "-fwrapv",
    "-pipe",
    "-w",
    "-fvisibility=hidden",
    "-fvisibility-inlines-hidden",
    "-Wno-unused-but-set-variable",
    "-O3",
    "-fPIC",
)


def locate_nuitka_headers() -> list[Path]:
    header_folders: list[Path] = []
    import nuitka

    nuitka_root = Path(nuitka.__file__).parent
    header_folders.append(nuitka_root / "build" / "static_src")
    header_folders.append(nuitka_root / "build" / "inline_copy" / "libbacktrace")
    header_folders.append(nuitka_root / "build" / "inline_copy" / "zlib")
    header_folders.append(nuitka_root / "build" / "include")

    return header_folders


def iterate_nuitka_c_sources(build_folder: str) -> Iterator[Path]:
    """
    Iterates over all C sources that should be compiled from the passed build fodler
    """
    root = Path(build_folder)
    for f in os.listdir(build_folder):
        if f.endswith(".c"):
            yield root / f

    static_src = root / "static_src"
    for f in os.listdir(static_src):
        if f.endswith(".c"):
            yield static_src / f


def compile_with_nuitka(
    path: str,
    no_follow_imports: bool = False,
    stdout: Stdout | None = None,
    include_modules: Iterable[str] | None = None,
    include_packages: Iterable[str] | None = None,
) -> str:
    """
    Compiles the module given by `path`.
    Follows imports by default, but can be disabled with `no_follow_imports`.
    """
    context = get_context()
    try:
        import nuitka

        # not using the import - just checking if it is available
        # as following logic would fail otherwise
        _ = nuitka
    except ImportError:
        raise ImportError(
            "Nuitka is not installed. Please install this package with nuitka extra: `pip install smelt[nuitka]`."
        )
    cmd = list(NUITKA_ENTRYPOINT)
    if not no_follow_imports:
        cmd.append("--follow-imports")
    cmd.append("--onefile")
    cmd.append(path)

    # handling special flags
    if include_modules:
        for mod in include_modules:
            cmd.append(f"--include-module={mod}")

    if include_packages:
        for package in include_packages:
            cmd.append(f"--include-package={package}")

    _logger.debug("Running %s", " ".join(cmd))

    cmd_trace = call_command(*cmd, printer=print)
    if context:
        context.add_trace(cmd_trace)
    if cmd_trace.exit_code != 0:
        raise RuntimeError(
            f"Nuitka failed with exitcode {cmd_trace.exit_code}: {' '.join(cmd)}"
        )

    expected_extension = ".exe" if sys.platform == "Windows" else ".bin"
    bin_path = os.path.basename(path).replace(".py", expected_extension)
    absolute_bin_path = os.path.join(os.getcwd(), bin_path)
    assert os.path.exists(absolute_bin_path), (
        f"Nuitka binary not found at {absolute_bin_path}"
    )
    return absolute_bin_path


def nuitkaify_module(
    module: NuitkaModule,
    path_solver: PathSolver | None = None,
    stdout: Stdout | None = None,
    include_modules: Iterable[str] | None = None,
    include_packages: Iterable[str] | None = None,
) -> GenericExtension:
    """
    Compiles the module given by `path`.
    Follows imports by default, but can be disabled with `no_follow_imports`.
    """
    path_solver = path_solver or PathSolver()
    context = get_context()
    src_path = module.source or path_solver.resolve_import_path(module.import_path)
    mod_filename = os.path.basename(src_path)
    cmd = list(NUITKA_ENTRYPOINT)
    cmd.append("--module")
    cmd.append(str(src_path))
    # TODO: the clean approach will to use the `--generate-c-only` since we want to
    # compile ourselves, however, when enabled nuitka generates the C code
    # but does not copy the static source files from its own package
    # These static files can be found in the .build folder when running a full build
    # but not when using generate-c-only.
    # The logic to find the ones that are required is not so trivial (nuitka only includes the ones)
    # that are actually needed, we we would have to tap into the scons stuff to extract them
    # ideal way would to do a dry run to extract all the build system data instead of trying
    # to reproduce it
    # As it now, nuitka will do the full compilation only for us to recompile again
    # cmd.append("--generate-c-only")

    # handling special flags
    if include_modules:
        for mod in include_modules:
            cmd.append(f"--include-module={mod}")

    if include_packages:
        for package in include_packages:
            cmd.append(f"--include-package={package}")

    _logger.debug("Running %s", " ".join(cmd))

    cmd_trace = call_command(*cmd, printer=print if stdout == "stdout" else None)
    if context:
        context.add_trace(cmd_trace)
    if cmd_trace.exit_code != 0:
        raise RuntimeError(
            f"Nuitka failed with exitcode {cmd_trace.exit_code}: {' '.join(cmd)}"
        )

    build_folder = mod_filename.replace(".py", ".build")
    assert os.path.exists(build_folder)
    c_sources = [str(src) for src in iterate_nuitka_c_sources(build_folder)]
    assert c_sources, (
        "Nuitka did not produce any C file or build folder path logic is incorrect"
    )
    header_sources = [str(f) for f in locate_nuitka_headers()]
    header_sources.append(build_folder)
    # patching build_definitions.h, as we don't need extensions
    open(os.path.join(build_folder, "build_definitions.h"), "w+").close()
    setup_tools_ext = Extension(
        name=mod_filename.replace(".py", ""),
        sources=c_sources,
        include_dirs=header_sources,
        define_macros=NUITKA_MACROS,
        libraries=["m", "dl", "z"],
        extra_compile_args=list(NUITKA_MINIMAL_FLAGS),
    )
    return GenericExtension(
        import_path=module.import_path,
        src_path=str(src_path),
        extension=setup_tools_ext,
        dest_folder=Path(src_path).parent,
    )
